import sys
from typing import Optional

from awsglue.context import GlueContext
from awsglue.utils import getResolvedOptions
from pyspark.sql import SparkSession

from capepy.aws.meta import Boto3Object


class EtlJob(Boto3Object):
    """An object for creating ETL Jobs for use in AWS Glue

    Attributes:
        spark_ctx: The PySpark session
        glue_ctx: The AWS Glue context
        logger: The logger for logging to AWS Glue
        parameters: A dictionary of parameters passed into the job
    """

    def __init__(self):
        """Constructor for instantiating a new Glue ETL Job"""
        super().__init__()
        self.spark_ctx = SparkSession.builder.getOrCreate()  # pyright: ignore
        self.glue_ctx = GlueContext(self.spark_ctx)
        self.logger = self.glue_ctx.get_logger()
        self.parameters = getResolvedOptions(
            sys.argv,
            [
                "RAW_BUCKET_NAME",
                "OBJECT_KEY",
                "CLEAN_BUCKET_NAME",
            ],
        )

    def get_raw_file(self):
        """Retrieve the raw file from S3 and return its contents as a byte string.

        Raises:
            Exception: If the raw file is unable to be successfully retrieved from S3.

        Returns:
            A byte string of the raw file contents
        """
        response = self.get_client("s3").get_object(
            Bucket=self.parameters["RAW_BUCKET_NAME"],
            Key=self.parameters["OBJECT_KEY"],
        )
        status = response.get("ResponseMetadata", {}).get("HTTPStatusCode")

        if status != 200:
            err = (
                f"ERROR - Could not get object {self.parameters['OBJECT_KEY']} from "
                f"bucket {self.parameters['RAW_BUCKET_NAME']}. ETL Cannot continue."
            )

            self.logger.error(err)

            # NOTE: need to properly handle exception stuff here, and we probably want
            #       this going somewhere very visible (e.g. SNS topic or a perpetual log
            #       as someone will need to be made aware)
            raise Exception(err)

        self.logger.info(
            f"Obtained object {self.parameters['OBJECT_KEY']} from bucket"
            f"{self.parameters['RAW_BUCKET_NAME']}"
        )

        return response.get("Body").read()

    def write_clean_file(self, clean_data, clean_key: Optional[str] = None):
        """Write data to a clean data file inside the clean S3 bucket as configured by the Glue ETL job.

        Args:
            clean_data (byes or seekable file-like object): Object data to be written to s3.
            clean_key: The prefix and filename for the new clean data file within the clean s3 bucket.

        Raises:
            Exception: If the clean data file is unable to be successfully put into s3.
        """
        response = self.get_client("s3").put_object(
            Bucket=self.parameters["CLEAN_BUCKET_NAME"],
            Key=clean_key,
            Body=clean_data,
        )

        status = response.get("ResponseMetadata", {}).get("HTTPStatusCode")

        if status != 200:
            err = (
                f"ERROR - Could not write transformed data object {clean_key} "
                f"to bucket {self.parameters['CLEAN_BUCKET_NAME']}. ETL Cannot continue."
            )

            self.logger.error(err)

            # NOTE: need to properly handle exception stuff here, and we probably
            #       want this going somewhere very visible (e.g. SNS topic or a
            #       perpetual log as someone will need to be made aware)
            raise Exception(err)

        self.logger.info(
            f"Transformed {self.parameters['RAW_BUCKET_NAME']}/{self.parameters['OBJECT_KEY']} and wrote result "
            f"to {self.parameters['CLEAN_BUCKET_NAME']}/{clean_key}"
        )



import json
from urllib.parse import unquote_plus


class Record(object):
    """An object for general records in AWS Lambda handlers.

    Attributes:
        raw: The raw record
    """

    def __init__(self, record):
        """Constructor for instantiating a new Lambda Record

        Args:
            record (object): A record from an AWS Lambda handler event
        """
        self.raw = record


class BucketNotificationRecord(Record):
    """An object for S3 bucket notification related records passed into AWS Lambda handlers.

    Attributes:
        bucket: The name of the bucket
        key: The key into the bucket if relevant
    """

    def __init__(self, record):
        """Constructor for instantiating a new record of S3 bucket information

        Args:
            record (object): An S3 bucket related record from an AWS Lambda handler event
        """
        super().__init__(record)
        self.bucket = self.raw["s3"]["bucket"]["name"]
        self.key = unquote_plus(
            self.raw["s3"]["object"]["key"], encoding="utf-8"
        )


class QueueRecord(Record):
    """An object for general records in AWS Lambda handlers.

    Attributes:
        raw: The raw record
    """

    def __init__(self, record):
        """Constructor for instantiating a new Lambda Record

        Args:
            record (object): A record from an AWS Lambda handler event
        """
        super().__init__(record)
        try:
            self.body = json.loads(self.raw["body"])
        except ValueError:
            self.body = self.raw["body"]
        self.id = self.raw["messageId"]


class EtlRecord(QueueRecord):
    """An object for ETL related records passed into AWS Lambda handlers.

    Attributes:
        job: The name of the ETL Job
        bucket: The name of the bucket
        key: The key into the bucket if relevant
    """

    def __init__(self, record):
        """Constructor for instantiating a new record of S3 bucket information

        Args:
            record (object): An S3 bucket related record from an AWS Lambda handler event
        """
        super().__init__(record)
        self.job = self.body["etl_job"]
        self.bucket = self.body["bucket"]
        self.key = self.body["key"]


class PipelineRecord(QueueRecord):
    """An object for pipeline records passed into AWS Lambda handlers.

    Attributes:
        name: The name of the analysis pipeline
        version: The version of the analysis pipeline
        parameters: A dictionary of parameters to pass to the analysis pipeline
    """

    def __init__(self, record):
        """Constructor for instantiating a new record of an analysis pipeline

        Args:
            record (object): An analysis pipeline related record from an AWS Lambda handler event
        """
        super().__init__(record)
        self.name = self.body["pipeline_name"]
        self.version = self.body["pipeline_version"]
        self.parameters = self.body["pipeline_parameters"]



import logging

import boto3


class Boto3Object(object):
    """Contains general resources needed by all AWS utilities for interacting with the boto3 library

    Attributes:
        logger: The logger for logging to AWS Glue
        clients: A dictionary of instantiated AWS clients indexed by the name of the AWS service.
        resources: A dictionary of instantiated AWS resources indexed by the name of the AWS service.
    """

    def __init__(self):
        """Constructor for initializing a Boto3Object"""
        self.logger = logging.getLogger(__name__)
        self.clients = {}
        self.resources = {}

    def get_client(self, service_name: str, **kwargs):
        """Get a client for the provided service, if one hasn't been created yet, set a new client.

        Args:
            service_name: The name of the service to retrieve a client for.
            **kwargs: Optional keyword arguments passed to `boto3.client()` if a new client needs to be set.

        Returns:
            The boto3 client.
        """
        if service_name not in self.clients:
            self.set_client(service_name, **kwargs)
        return self.clients[service_name]

    def set_client(self, service_name: str, **kwargs):
        """Set a new client for the provided service.

        Args:
            service_name: The name of the service to instantiate a new client
            for.
            **kwargs: Optional keyword arguments passed to `boto3.client()`.
        """
        self.clients[service_name] = boto3.client(
            service_name, **kwargs  # pyright: ignore
        )

    def get_resource(self, service_name: str, **kwargs):
        """Get a resource for the provided service, if one hasn't been created yet, set a new resource.

        Args:
            service_name: The name of the service to retrieve a resource for.
            **kwargs: Optional keyword arguments passed to `boto3.resource()` if a new resource needs to be set.

        Returns:
            The boto3 client.
        """
        if service_name not in self.resources:
            self.set_resource(service_name, **kwargs)
        return self.resources[service_name]

    def set_resource(self, service_name: str, **kwargs):
        """Set a new resource for the provided service.

        Args:
            service_name: The name of the service to instantiate a new resource for.
            **kwargs: Optional keyword arguments passed to `boto3.resource()`.
        """
        self.resources[service_name] = boto3.resource(
            service_name, **kwargs  # pyright: ignore
        )



from botocore.exceptions import ClientError


def decode_error(err: ClientError):
    """Decode a client error message from AWS

    Args:
        err: The ClientError to parse out the error code and message if they are
        available.

    Returns:
        A tuple (code, message) where code is a string containing the error
        code, and message is a string containing the entire error message.
    """
    code, message = "Unknown", "Unknown"
    if "Error" in err.response:
        error = err.response["Error"]
        if "Code" in error:
            code = error["Code"]
        if "Message" in error:
            message = error["Message"]
    return code, message



import os

from botocore.exceptions import ClientError

from capepy.aws.meta import Boto3Object
from capepy.aws.utils import decode_error


class Table(Boto3Object):
    """An object for working with specific DynamoDB Table structures in the CAPE system.

    Attributes:
        name: the name of the table
        table: the table object retrieved with the boto3 dynamodb resource
    """

    def __init__(self, table_name):
        """Constructor for instantiating the DynamoDB table object. Assumes DDB_REGION is set in the environment.

        Args:
            table_name: The name of the table to retrieve from DynamoDB.
        """
        super().__init__()
        ddb = self.get_resource("dynamodb", region_name=os.getenv("DDB_REGION"))
        self.name = table_name

        try:
            self.table = ddb.Table(table_name)
            self.table.load()
        except ClientError as err:
            code, message = decode_error(err)

            if code == "ResourceNotFoundException":
                msg = (
                    f"CAPE DAP registry DynamoDB table ({table_name}) could not"
                    f"be found: {code} {message}",
                )
            else:
                msg = (
                    f"Error trying to access CAPE DynamoDB table ({table_name}): "
                    f"{code} {message}",
                )

            self.logger.error(msg)
            raise err

    def get_item(self, key):
        """Retrieve an item from the loaded table.

        Args:
            key: The key of the entry to retrieve from the table.

        Returns:
            The item value from the DyamoDB table.
        """
        ret = None
        try:
            ret = self.table.get_item(Key=key)["Item"]

        except ClientError as err:
            code, message = decode_error(err)

            self.logger.error(
                f"Couldn't get DynamoDB entry for key '{key}' in table {self.name}. "
                f"{code} {message}"
            )

        return ret


class PipelineTable(Table):
    """A DynamoDB table with specific structure for organizing analysis pipelines."""

    def __init__(self, table_name=None):
        """Constructor to fetch and initialize a DynamoDB analysis pipeline table.

        Args:
            table_name: The name of the pipeline table to retrieve from DynamoDB. Defaults to $DAP_REG_DDB_TABLE
        """
        if table_name is None:
            table_name = os.getenv("DAP_REG_DDB_TABLE")
        super().__init__(table_name)

    def get_pipeline(self, pipeline_name, pipeline_version):
        """Retrieve a specific pipeline from the table.

        Args:
            pipeline_name: The name of the pipeline.
            pipeline_version: The version of the pipeline.

        Returns:
            The retrieved pipeline item.
        """
        key = {"pipeline_name": pipeline_name}
        if pipeline_version is not None:
            key["version"] = pipeline_version
        return self.get_item(key)


class EtlTable(Table):
    """A DynamoDB table with specific structure for organizing ETL jobs."""

    def __init__(self, table_name=None):
        """Constructor to fetch and initialize a DynamoDB ETL job table.

        Args:
            table_name: The name of the ETL table to retrieve from DynamoDB. Defaults to $ETL_ATTRS_DDB_TABLE
        """
        if table_name is None:
            table_name = os.getenv("ETL_ATTRS_DDB_TABLE")
        super().__init__(table_name)

    def get_etls(self, bucket_name, prefix):
        """Retrieve a specific ETLs from the table.

        Args:
            bucket_name: The name of the s3 bucket to get ETL jobs for.
            prefix: The required prefix to check for ETL jobs.

        Returns:
            The ETL Jobs triggered by the given bucket name and prefix.
        """
        return self.get_item({"bucket_name": bucket_name, "prefix": prefix})
